@@ -1,6 +1,7 @@
 import enum
 
-from zigpy.types import uint8_t
+import zigpy.config
+import zigpy.types as t
 
 
 class Repr:
@@ -17,20 +18,36 @@ class Timeouts:
     default = 10000
 
 
-class ZnpVersion(uint8_t, enum.Enum):
+class AddressMode(t.uint8_t, enum.Enum):
+    ADDR_NOT_PRESENT = 0
+    ADDR_GROUP = 1
+    ADDR_16BIT = 2
+    ADDR_64BIT = 3
+    ADDR_BROADCAST = 15
+
+
+class LedMode(t.uint8_t, enum.Enum):
+    Off = 0
+    On = 1
+    Blink = 2
+    Flash = 3
+    Toggle = 4
+
+
+class ZnpVersion(t.uint8_t, enum.Enum):
     zStack12 = 0
     zStack3x0 = 1
     zStack30x = 2
 
 
-class CommandType(uint8_t, enum.Enum):
+class CommandType(t.uint8_t, enum.Enum):
     POLL = 0
     SREQ = 1
     AREQ = 2
     SRSP = 3
 
 
-class Subsystem(uint8_t, enum.Enum):
+class Subsystem(t.uint8_t, enum.Enum):
     RESERVED = 0
     SYS = 1
     MAC = 2
@@ -45,7 +62,7 @@ class Subsystem(uint8_t, enum.Enum):
     GREENPOWER = 21
 
 
-class ParameterType(uint8_t, enum.Enum):
+class ParameterType(t.uint8_t, enum.Enum):
     UINT8 = 0
     UINT16 = 1
     UINT32 = 2
@@ -81,10 +98,21 @@ class ParameterType(uint8_t, enum.Enum):
         )
 
 
-class NetworkOptions:
-    def __init__(self) -> None:
+class NetworkOptions(Repr):
+    networkKey: t.KeyData
+    panID: t.PanId
+    extendedPanID: t.ExtendedPanId
+    channelList: t.Channels
+
+    def __init__(self, config: zigpy.config.SCHEMA_NETWORK) -> None:
         self.networkKeyDistribute = False
-        self.networkKey = [1, 3, 5, 7, 9, 11, 13, 15, 0, 2, 4, 6, 8, 10, 12, 13]
-        self.panID = 0x1A62
-        self.extendedPanID = [0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD]
-        self.channelList = [11]
+        self.networkKey = config[zigpy.config.CONF_NWK_KEY] or (
+            zigpy.config.cv_key([1, 3, 5, 7, 9, 11, 13, 15, 0, 2, 4, 6, 8, 10, 12, 13])
+        )
+        self.panID = config[zigpy.config.CONF_NWK_PAN_ID] or t.PanId(0x1A62)
+        self.extendedPanID = config[zigpy.config.CONF_NWK_EXTENDED_PAN_ID] or (
+            t.ExtendedPanId([0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD])
+        )
+        self.channelList = (
+            config[zigpy.config.CONF_NWK_CHANNELS] or t.Channels.from_channel_list([11])
+        )
