@@ -1,20 +1,22 @@
 import asyncio
 import logging
+from asyncio.locks import Semaphore
 from typing import Any, Dict, Optional
 
 import zigpy.application
 import zigpy.config
 import zigpy.device
-from zigpy.profiles import zha
 import zigpy.types
 import zigpy.util
+from zigpy.profiles import zha
+from zigpy.types import BroadcastAddress
 from zigpy.zdo.types import ZDOCmd
 
 from zigpy_cc import __version__, types as t
 from zigpy_cc.api import API
-from zigpy_cc.config import CONF_DEVICE, CONFIG_SCHEMA
+from zigpy_cc.config import CONF_DEVICE, CONFIG_SCHEMA, SCHEMA_DEVICE
 from zigpy_cc.exception import TODO, CommandError
-from zigpy_cc.types import NetworkOptions, Subsystem, ZnpVersion
+from zigpy_cc.types import NetworkOptions, Subsystem, ZnpVersion, LedMode, AddressMode
 from zigpy_cc.zigbee.start_znp import start_znp
 from zigpy_cc.zpi_object import ZpiObject
 
@@ -29,6 +31,7 @@ REQUESTS = {
     "ieeeAddrReq": (ZDOCmd.IEEE_addr_req, 0),
     "matchDescReq": (ZDOCmd.Match_Desc_req, 2),
     "endDeviceAnnceInd": (ZDOCmd.Device_annce, 2),
+    "mgmtLqiRsp": (ZDOCmd.Mgmt_Lqi_rsp, 2),
     "mgmtPermitJoinReq": (ZDOCmd.Mgmt_Permit_Joining_req, 3),
     "mgmtPermitJoinRsp": (ZDOCmd.Mgmt_Permit_Joining_rsp, 2),
     "nodeDescRsp": (ZDOCmd.Node_Desc_rsp, 2),
@@ -38,10 +41,10 @@ REQUESTS = {
 }
 
 IGNORED = (
-    # "activeEpRsp",
+    "bdbComissioningNotifcation",
     "dataConfirm",
     "leaveInd",
-    # "mgmtPermitJoinRsp",
+    "resetInd",
     "srcRtgInd",
     "stateChangeInd",
     "tcDeviceInd",
@@ -49,8 +52,12 @@ IGNORED = (
 
 
 class ControllerApplication(zigpy.application.ControllerApplication):
+    _semaphore: Semaphore
     _api: Optional[API]
     SCHEMA = CONFIG_SCHEMA
+    SCHEMA_DEVICE = SCHEMA_DEVICE
+
+    probe = API.probe
 
     def __init__(self, config: Dict[str, Any]):
         super().__init__(config=zigpy.config.ZIGPY_SCHEMA(config))
@@ -81,11 +88,23 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     async def startup(self, auto_form=False):
         """Perform a complete application startup"""
 
-        LOGGER.debug("Starting zigpy-cc version: %s", __version__)
+        LOGGER.info("Starting zigpy-cc version: %s", __version__)
         self._api = await API.new(self, self._config[CONF_DEVICE])
+
+        try:
+            await self._api.request(Subsystem.SYS, "ping", {"capabilities": 1})
+        except CommandError as e:
+            raise Exception("Failed to connect to the adapter(%s)", e)
+
         self.version = await self._api.version()
+
+        concurrent = 16 if self.version["product"] == ZnpVersion.zStack3x0 else 2
+        LOGGER.debug("Adapter concurrent: %d", concurrent)
+
+        self._semaphore = asyncio.Semaphore(concurrent)
+
         ver = ZnpVersion(self.version["product"]).name
-        LOGGER.debug("Detected znp version '%s' (%s)", ver, self.version)
+        LOGGER.info("Detected znp version '%s' (%s)", ver, self.version)
 
         if auto_form:
             await self.form_network()
@@ -106,12 +125,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
     async def form_network(self, channel=15, pan_id=None, extended_pan_id=None):
         LOGGER.info("Forming network")
-        options = NetworkOptions()
+        LOGGER.debug("Config: %s", self.config)
+        options = NetworkOptions(self.config[zigpy.config.CONF_NWK])
+        LOGGER.debug("NetworkOptions: %s", options)
         backupPath = ""
         status = await start_znp(
-            self._api, self.version["product"], options, backupPath
+            self._api, self.version["product"], options, 0x0B84, backupPath
         )
-        LOGGER.debug("ZNP started, status: %s", status)
+        LOGGER.info("ZNP started, status: %s", status)
+
+        self.set_led(LedMode.Off)
 
     async def mrequest(
         self,
@@ -141,7 +164,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         :returns: return a tuple of a status and an error_message. Original requestor
                   has more context to provide a more meaningful error message
         """
-        req_id = self.get_sequence()
         LOGGER.debug(
             "multicast %s",
             (
@@ -157,10 +179,28 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         )
         try:
             obj = ZpiObject.from_cluster(
-                group_id, profile, cluster, src_ep, src_ep, sequence, data, req_id
+                group_id,
+                profile,
+                cluster,
+                src_ep or 1,
+                0xFF,
+                sequence,
+                data,
+                addr_mode=AddressMode.ADDR_GROUP,
             )
+            waiter_id = None
+            waiter = self._api.create_response_waiter(obj, sequence)
+            if waiter:
+                waiter_id = waiter.id
 
-            await self._api.request_raw(obj)
+            async with self._semaphore:
+                await self._api.request_raw(obj, waiter_id)
+                """
+                As a group command is not confirmed and thus immediately returns
+                (contrary to network address requests) we will give the
+                command some time to 'settle' in the network.
+                """
+                await asyncio.sleep(0.2)
 
         except CommandError as ex:
             return ex.status, "Couldn't enqueue send data multicast: {}".format(ex)
@@ -180,7 +220,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         expect_reply=True,
         use_ieee=False,
     ):
-        req_id = self.get_sequence()
         LOGGER.debug(
             "request %s",
             (
@@ -198,12 +237,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         try:
             obj = ZpiObject.from_cluster(
-                device.nwk, profile, cluster, src_ep, dst_ep, sequence, data, req_id
+                device.nwk, profile, cluster, src_ep, dst_ep, sequence, data
             )
+            waiter_id = None
             if expect_reply:
-                self._api.create_response_waiter(obj, sequence)
+                waiter = self._api.create_response_waiter(obj, sequence)
+                if waiter:
+                    waiter_id = waiter.id
 
-            await self._api.request_raw(obj)
+            async with self._semaphore:
+                await self._api.request_raw(obj, waiter_id)
 
         except CommandError as ex:
             return ex.status, "Couldn't enqueue send data request: {}".format(ex)
@@ -222,7 +265,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         data,
         broadcast_address=zigpy.types.BroadcastAddress.RX_ON_WHEN_IDLE,
     ):
-        req_id = self.get_sequence()
         LOGGER.debug(
             "broadcast %s",
             (
@@ -246,11 +288,18 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 dst_ep,
                 sequence,
                 data,
-                req_id,
                 radius=radius,
+                addr_mode=AddressMode.ADDR_16BIT,
             )
 
-            await self._api.request_raw(obj)
+            async with self._semaphore:
+                await self._api.request_raw(obj)
+                """
+                As a broadcast command is not confirmed and thus immediately returns
+                (contrary to network address requests) we will give the
+                command some time to 'settle' in the network.
+                """
+                await asyncio.sleep(0.2)
 
         except CommandError as ex:
             return (
@@ -263,24 +312,18 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     async def permit_ncp(self, time_s=60):
         assert 0 <= time_s <= 254
         payload = {
-            "addrmode": 0x0F,
-            "dstaddr": 0xFFFC,
+            "addrmode": AddressMode.ADDR_BROADCAST,
+            "dstaddr": BroadcastAddress.ALL_ROUTERS_AND_COORDINATOR,
             "duration": time_s,
             "tcsignificance": 0,
         }
-        await self._api.request(Subsystem.ZDO, "mgmtPermitJoinReq", payload)
+        async with self._semaphore:
+            await self._api.request(Subsystem.ZDO, "mgmtPermitJoinReq", payload)
 
     def handle_znp(self, obj: ZpiObject):
-        if obj.type != t.CommandType.AREQ:
+        if obj.command_type != t.CommandType.AREQ:
             return
 
-        # if obj.subsystem == t.Subsystem.ZDO and obj.command == 'tcDeviceInd':
-        #     nwk = obj.payload['nwkaddr']
-        #     rest = obj.payload['extaddr'][2:].encode("ascii")
-        #     ieee, _ = zigpy.types.EUI64.deserialize(rest)
-        #     LOGGER.info("New device joined: 0x%04x, %s", nwk, ieee)
-        #     self.handle_join(nwk, ieee, obj.payload['parentaddr'])
-
         frame = obj.to_unpi_frame()
 
         nwk = obj.payload["srcaddr"] if "srcaddr" in obj.payload else None
@@ -295,26 +338,22 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             nwk = obj.payload["nwkaddr"]
             LOGGER.info("New device joined: 0x%04x, %s", nwk, ieee)
             self.handle_join(nwk, ieee, 0)
-            # TODO TEST
             obj.sequence = 0
 
-        # TODO bindRsp
-
         if obj.command in IGNORED:
-            LOGGER.debug("message ignored: %s", obj.command)
             return
 
+        if obj.subsystem == t.Subsystem.ZDO and obj.command == "mgmtPermitJoinRsp":
+            self.set_led(LedMode.On)
+
         if obj.subsystem == t.Subsystem.ZDO and obj.command == "permitJoinInd":
-            # if obj.payload["duration"] == 0:
-            #     loop = asyncio.get_event_loop()
-            #     loop.create_task(self.permit_ncp())
+            self.set_led(LedMode.Off if obj.payload["duration"] == 0 else LedMode.On)
             return
 
         if obj.subsystem == t.Subsystem.ZDO and obj.command in REQUESTS:
             if obj.sequence is None:
-                LOGGER.warning("missing tsn from %s, maybe not a reply", obj.command)
                 return
-            LOGGER.info("REPLY for %d %s", obj.sequence, obj.command)
+            LOGGER.debug("REPLY for %d %s", obj.sequence, obj.command)
             cluster_id, prefix_length = REQUESTS[obj.command]
             tsn = bytes([obj.sequence])
             data = tsn + frame.data[prefix_length:]
@@ -331,7 +370,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         else:
             LOGGER.warning(
-                "Unhandled message: %s %s", t.Subsystem(obj.subsystem), obj.command
+                "Unhandled message: %s %s %s",
+                t.CommandType(obj.command_type),
+                t.Subsystem(obj.subsystem),
+                obj.command,
             )
             return
 
@@ -347,13 +389,29 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             return
 
         device.radio_details(lqi, rssi)
-        # if obj.subsystem == t.Subsystem.ZDO:
-        #     pass
+
         LOGGER.info("handle_message %s", obj.command)
         self.handle_message(device, profile_id, cluster_id, src_ep, dst_ep, data)
 
+    def set_led(self, mode: LedMode):
+        if self.version["product"] != ZnpVersion.zStack3x0:
+            try:
+                loop = asyncio.get_event_loop()
+                loop.create_task(
+                    self._api.request(
+                        Subsystem.UTIL, "ledControl", {"ledid": 3, "mode": mode}
+                    )
+                )
+            except Exception as ex:
+                LOGGER.warning("Can't set LED: %s", ex)
+
 
 class Coordinator(zigpy.device.Device):
+    """
+    todo add endpoints?
+    @see zStackAdapter.ts - getCoordinator
+    """
+
     @property
     def manufacturer(self):
         return "Texas Instruments"
